import logging

import requests

from abc import ABC, abstractmethod


# we could reuse the reporters instead
from requests import JSONDecodeError

from kensu.utils.exceptions import SdkError


class AbstractSDK(ABC):
    @abstractmethod
    def get_cookie(self):
        pass

    @abstractmethod
    def get_lineages_in_project(self, project, process, env, code_version):
        pass

    @abstractmethod
    def create_rule(self, lds_id, lineage_id, project_id, process_id, env_name, field_name, fun):
        pass

    @abstractmethod
    def update_rule(self, predicate, fun):
        pass

    @abstractmethod
    def get_rules(self):
        pass

    @abstractmethod
    def get_all_rules_for_ds(self, ds_id):
        pass

    @abstractmethod
    def get_rules_for_ds_in_project(self, ds_id, lineage_id, project_id, env):
        pass

    @abstractmethod
    def get_datasources_in_logical(self, logical):
        pass

    @abstractmethod
    def get_datasource(self, dsId):
        pass

    @abstractmethod
    def get_latest_datasource_in_logical(self, logical, n=-1):
        pass

    @abstractmethod
    def get_schema(self, schema_id):
        pass

    @abstractmethod
    def get_latest_schema_in_datasource(self, ds):
        pass

    @abstractmethod
    def get_latest_schema_in_logical(self, logical,n=-1):
        pass

    @abstractmethod
    def get_latest_stats_for_ds(self, projectId, env, linId, dsId):
        pass

    @abstractmethod
    def get_logical_ds_name_from_ds(self, dsId):
        pass

    @abstractmethod
    def get_datasources(self):
        pass


class DoNothingSDK(ABC):
    def __init__(self):
        pass

    def get_cookie(self):
        pass
    
    def get_lineages_in_project(self, project, process, env, code_version):
        pass
    
    def create_rule(self, lds_id, lineage_id, project_id, process_id, env_name, field_name, fun):
        pass
    
    def update_rule(self, predicate, fun):
        pass
    
    def get_rules(self):
        pass

    def get_all_rules_for_ds(self, ds_id):
        pass
    
    def get_rules_for_ds_in_project(self, ds_id, lineage_id, project_id, env):
        pass
    
    def get_datasources_in_logical(self, logical):
        pass
    
    def get_datasource(self, dsId):
        pass
    
    def get_latest_datasource_in_logical(self, logical, n=-1):
        pass
    
    def get_schema(self, schema_id):
        pass
    
    def get_latest_schema_in_datasource(self, ds):
        pass
    
    def get_latest_schema_in_logical(self, logical, n=-1):
        pass
    
    def get_latest_stats_for_ds(self, projectId, env, linId, dsId):
        pass
    
    def get_logical_ds_name_from_ds(self, dsId):
        pass

    def get_datasources(self):
        pass


class SDK(AbstractSDK):
    def __init__(self, url, sdk_token, verify_ssl):
        self.url = url
        self.PAT = sdk_token
        self.cookie_url = self.url + '/api/auth/callback?client_name=ExternalAppTokenClient'
        self.cookie_header = {'X-External-App-Token': self.PAT}
        self.verify_ssl = verify_ssl
        self.cookie = self.get_cookie()  # FIXME: get_cookie called only once in __init__()! thus it might get expired!

    def get_cookie(self):
        session = requests.Session()
        response = session.post(url=self.cookie_url, headers=self.cookie_header, verify=self.verify_ssl)
        cookie = session.cookies
        return cookie

    def requests_get_json(self, uri_suffix):
        # FIXME: proper URI concat
        # FIXME: verify if this update the cookie jar?
        # FIXME: what if cookie expired?
        resp = requests.get(self.url + uri_suffix, cookies=self.cookie, verify=self.verify_ssl)
        if not (200 <= resp.status_code <= 299):
            msg = f"Failed to query Kensu SDK for uri={uri_suffix} status={resp.status_code}:\n{resp.text}"
            e = SdkError(msg)
            logging.warning(msg, e)
            raise SdkError(e)
        try:
            return resp.json()
        except JSONDecodeError as e:
            logging.warning(f"Unable to decode Kensu SDK response for uri={uri_suffix}:\n{v}", e)
            raise SdkError(e)


    def get_lineages_in_project(self, project, process, env, code_version):
        # FIXME: use proper URLencode
        uri = "/business/api/views/v1/project-catalog/process/data-flow?projectId=%s&processId=%s&logical=true&environment=%s&codeVersionId=%s" % (project,process,env,code_version)
        return self.requests_get_json(uri)

    def create_rule(self, lds_id, lineage_id=None, project_id=None, process_id=None, env_name=None, field_name=None, fun=None, context="DATA_STATS"):
        uri = "/business/api/v1/predicates"
        payload = None
        if context == "DATA_STATS":
            payload = {
                "context": "DATA_STATS",
                "datasourceId": lds_id,
                "lineageId": lineage_id,
                "projectId": project_id,
                "processId": process_id,
                "environment": env_name,
                "isLogical": True,
                "fieldName": field_name,
                "functionName": fun["name"],
                "arguments": fun["arguments"]
            }
        elif context == "LOGICAL_DATA_SOURCE":
            payload = {
                "context": "LOGICAL_DATA_SOURCE",
                "datasourceId": lds_id,
                "environment": "",
                "fieldName": field_name,
                "functionName": fun["name"],
                "arguments": fun["arguments"]
            }

        if payload:
            return requests.post(self.url + uri, json=payload, cookies=self.cookie, verify=self.verify_ssl).json()

    def update_rule(self, predicate, fun):
        uri = "/business/api/v1/predicates/%s" % predicate

        payload = {"functionName": fun["name"],
                   "arguments": fun["arguments"]}

        v = requests.put(self.url + uri, json=payload, cookies=self.cookie, verify=self.verify_ssl)
        return None

    def get_rules(self):
        uri = "/business/api/views/v1/predicate-catalog"
        return self.requests_get_json(uri)

    def get_rules_for_ds_in_project(self, ds_id, lineage_id, project_id, env):
        uri = "/business/api/v1/performance/data/%s/%s?projectId=%s&logical=true&environment=%s" % (ds_id,lineage_id,project_id,env)
        return self.requests_get_json(uri)

    def get_all_rules_for_ds(self,ds_id):
        uri = "/business/api/v1/predicates?logical_data_source_id=%s&context=LOGICAL_DATA_SOURCE" % (ds_id)
        return self.requests_get_json(uri)

    def get_datasources_in_logical(self, logical):
        # FIXME: 404
        return self.requests_get_json("/api/services/v1/experimental/datasources/in-logical/%s" % logical)

    def get_datasource(self, dsId):
        return self.requests_get_json("/api/services/v1/resources/datasource/%s" % dsId)

    def get_latest_datasource_in_logical(self, logical, n=-1):
        js = self.get_datasources_in_logical(logical)
        sorted_js = (sorted((i for i in js), key=lambda k: k['timestamp']))

        if len(sorted_js)>=abs(n):
            uuid = sorted_js[n]["uuid"]
            ds = self.get_datasource(uuid)
            return ds
        else:
            return None

    def get_schema(self, schema_id):
        return self.requests_get_json("/api/services/v1/resources/schema/%s" % schema_id)

    def get_latest_schema_in_datasource(self, ds):
        if ds:
            schemas = ds['schemas']
            schema_uuid = (max((i for i in schemas), key=lambda k: k['timestamp']))['schemaId']
            schema = self.get_schema(schema_uuid)
            return {x["columnName"]:x["columnType"] for x in schema['schema']}
        else:
            return None

    def get_latest_schema_in_logical(self, logical, n=-1):
        ds = self.get_latest_datasource_in_logical(logical, n)
        schema = self.get_latest_schema_in_datasource(ds)
        return schema

    def get_latest_stats_for_ds(self, projectId, env, linId, dsId):
        resp_json = self.requests_get_json("/business/api/v1/performance/data/%s/%s?projectId=%s&logical=false&environment=%s" % (dsId,linId,projectId,env))
        stats_json = sorted(resp_json['data']['stats'], key=lambda k: k['timestamp'])[-1]
        return stats_json['stats']

    def get_logical_ds_name_from_ds(self, dsId):
        ds = self.requests_get_json("/business/api/v1/datasources/%s" % dsId)
        return ds["data"]["logicalDatasource"]["name"]

    def get_datasources(self):
        return self.requests_get_json("/api/services/v1/resources/datasources")



import sqlparse
from google.cloud.bigquery import client, Dataset
import logging

from kensu.client import DataSourcePK, Schema, SchemaPK
from kensu.utils.helpers import to_datasource

logger = logging.getLogger(__name__)

from .job.offline_parser import BqOfflineParser
from .job.query import QueryJob
from google.cloud.bigquery.retry import DEFAULT_RETRY
from kensu.utils.kensu_provider import KensuProvider
from .job.remote_parser import BqRemoteParser
import google.cloud.bigquery as bq


class Client(client.Client):

    def query(
        self,
        query,
        job_config=None,
        job_id=None,
        job_id_prefix=None,
        location=None,
        project=None,
        retry=DEFAULT_RETRY,
        timeout=None
    ):
        gclient = super(Client, self)
        j = gclient.query(query,
        job_config,
        job_id,
        job_id_prefix,
        location,
        project,
        retry,
        timeout)
        kensu = KensuProvider().instance()
        kensu.data_collectors['BigQuery'] = gclient

        queries = query

        #For succession of SELECT queries, only the last one return a result
        for query in queries.split(";"):
            normalized_query = query.lower().replace(" ", "").replace("\n", "")
            if normalized_query.startswith("insertinto") or normalized_query.startswith("merge") :
                kensu = KensuProvider().instance()
                client = kensu.data_collectors['BigQuery']
                import sqlparse
                q = sqlparse.parse(query)
                d = BqOfflineParser.find_sql_identifiers(q[0].tokens).__next__()
                table = d.value.replace('`', '')
                ds_data = table.split('.')
                if project is not None and len(ds_data) == 2:
                    ds_data = [project] + ds_data
                if len(ds_data) == 3:
                    ds_data_corr = ds_data
                    ds_data = []
                    for element in ds_data_corr:
                        ds_data.append(element.split(' ')[0])
                    ds = ".".join(ds_data[0:2])
                    destination = client.create_dataset(Dataset(ds), timeout=30, exists_ok=True).table(
                        ds_data[2])
                    if isinstance(destination, bq.TableReference):
                        dest = client.get_table(destination)
                    db_metadata, table_id_to_bqtable, table_infos = BqOfflineParser.get_referenced_tables_metadata(
                        kensu=kensu,
                        client=client,
                        query=query.replace("\n", " ")
                    )

                    try:
                        #TODO: Use sqlparse
                        bq_lineage = BqRemoteParser.parse(
                            kensu=kensu,
                            client=client,
                            query=query,
                            db_metadata=db_metadata,
                            table_id_to_bqtable=table_id_to_bqtable)
                    except:
                        bq_lineage = BqOfflineParser.fallback_lineage(kensu, table_infos, dest)

                    db_metadata_out, table_id_to_bqtable_out, table_infos_out = BqOfflineParser.get_referenced_tables_metadata(
                        kensu=kensu,
                        client=client,
                        table=dest)

                    table_infos_out[0][1]._report()
                    bq_lineage.report(
                        ksu=kensu,
                        df_result=table_infos_out[0][2],
                        operation_type='BigQuery SQL result',
                        report_output=True,
                        # FIXME: how to know when in mem or when bigquery://projects/psyched-freedom-306508/datasets/_b63f45da1cafbd073e5c2770447d963532ac43ec/tables/anonc79d9038a13ab2dbe40064636b0aceedc62b5d69
                        register_output_orig_data=True  # FIXME? when do we need this? INSERT INTO?
                    )

                    from kensu.google.cloud.bigquery.job.bigquery_stats import compute_bigquery_stats
                    try:
                        import re
                        if query.lower().replace(" ", "").replace("\n", "").startswith("insertinto"):
                            index_select = query.lower().index("select")
                            query_without_insert = query[index_select:]

                        elif query.lower().replace(" ", "").replace("\n", "").startswith("merge"):
                            # query_without_insert = re.findall('\(.*\)', query)[0]
                            import sqlparse
                            query_without_insert = list(sqlparse.parse(query.replace("\n", " "))[0].get_sublists())[1].value
                            query_without_insert = re.findall('\(.*\)', query_without_insert)[0]

                        else:
                            query_without_insert = query
                        output_stats = compute_bigquery_stats(table_ref=destination, table = client.get_table(destination), client = client,query = query_without_insert)
                    except:
                        logger.debug(f"Unable to compute stats for table {str(table)}")
                        output_stats=None
                    kensu.real_schema_df[table_infos_out[0][2].to_guid()] = output_stats

                    kensu.report_with_mapping()

        return QueryJob.patch(j)


    def load_table_from_uri(
        self,
        source_uris,
        destination,
        job_id = None,
        job_id_prefix = None,
        location = None,
        project = None,
        job_config = None,
        retry = DEFAULT_RETRY,
        timeout = None
    ):
        gclient = super(Client, self)
        j = gclient.load_table_from_uri(source_uris,
        destination,
        job_id,
        job_id_prefix,
        location,
        project,
        job_config,
        retry,
        timeout)

        block = j.result()

        kensu = KensuProvider().instance()
        kensu.data_collectors['BigQuery'] = gclient

        dest_metadata, table_id_to_bqtable, table_infos = BqOfflineParser.get_referenced_tables_metadata(
            kensu=kensu,
            client=gclient,
            table=destination)
        ds = table_infos[0][1]._report()
        sc = table_infos[0][2]._report()
        sc_guid = sc.to_guid()
        kensu.real_schema_df[sc_guid] = None
        in_schema = sc


        if isinstance(source_uris, list):
            input_ds = source_uris
        else:
            input_ds = [source_uris]

        list_schemas = []
        for ds_uri_location in input_ds:
            #TODO Extractor?
            logical_naming = kensu.logical_naming
            location = ds_uri_location
            ds_pk = DataSourcePK(location=location, physical_location_ref=kensu.default_physical_location_ref)
            format = ds_uri_location.split(".")[-1] if "." in ds_uri_location else "Unknown"
            name = ('/').join(location.split('/')[-2:])
            ds_in = to_datasource(ds_pk, format, location, logical_naming, name)._report()
            ds_in_guid = ds_in.to_guid()
            in_schema.pk.data_source_ref.by_guid = ds_in_guid
            schema_in = Schema(name = "Schema:"+name, pk=in_schema.pk)._report()
            list_schemas.append(schema_in)
            #TODO replace none
            kensu.real_schema_df[schema_in.to_guid()] = None
            for col in [field.name for field in sc.pk.fields]:
                kensu.add_dependencies_mapping(sc_guid,col,schema_in.to_guid(),col,"load")

        kensu.report_with_mapping()

        return j



# coding: utf-8

"""
    

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    OpenAPI spec version: beta
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from pprint import pformat

from six import iteritems


class SchemaLineageDependencyDef(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'from_schema_ref': 'SchemaRef',
        'to_schema_ref': 'SchemaRef',
        'column_data_dependencies': 'dict(str, list[str])',
        'column_control_dependencies': 'dict(str, list[str])'
    }

    attribute_map = {
        'from_schema_ref': 'fromSchemaRef',
        'to_schema_ref': 'toSchemaRef',
        'column_data_dependencies': 'columnDataDependencies',
        'column_control_dependencies': 'columnControlDependencies'
    }

    def __init__(self, from_schema_ref=None, to_schema_ref=None, column_data_dependencies=None, column_control_dependencies=None):
        """
        SchemaLineageDependencyDef - a model defined in Swagger
        """

        self._from_schema_ref = None
        self._to_schema_ref = None
        self._column_data_dependencies = None
        self._column_control_dependencies = None

        self.from_schema_ref = from_schema_ref
        self.to_schema_ref = to_schema_ref
        if column_data_dependencies is not None:
          self.column_data_dependencies = column_data_dependencies
        if column_control_dependencies is not None:
          self.column_control_dependencies = column_control_dependencies

    @property
    def from_schema_ref(self):
        """
        Gets the from_schema_ref of this SchemaLineageDependencyDef.

        :return: The from_schema_ref of this SchemaLineageDependencyDef.
        :rtype: SchemaRef
        """
        return self._from_schema_ref

    @from_schema_ref.setter
    def from_schema_ref(self, from_schema_ref):
        """
        Sets the from_schema_ref of this SchemaLineageDependencyDef.

        :param from_schema_ref: The from_schema_ref of this SchemaLineageDependencyDef.
        :type: SchemaRef
        """
        if from_schema_ref is None:
            raise ValueError("Invalid value for `from_schema_ref`, must not be `None`")

        self._from_schema_ref = from_schema_ref

    @property
    def to_schema_ref(self):
        """
        Gets the to_schema_ref of this SchemaLineageDependencyDef.

        :return: The to_schema_ref of this SchemaLineageDependencyDef.
        :rtype: SchemaRef
        """
        return self._to_schema_ref

    @to_schema_ref.setter
    def to_schema_ref(self, to_schema_ref):
        """
        Sets the to_schema_ref of this SchemaLineageDependencyDef.

        :param to_schema_ref: The to_schema_ref of this SchemaLineageDependencyDef.
        :type: SchemaRef
        """
        if to_schema_ref is None:
            raise ValueError("Invalid value for `to_schema_ref`, must not be `None`")

        self._to_schema_ref = to_schema_ref

    @property
    def column_data_dependencies(self):
        """
        Gets the column_data_dependencies of this SchemaLineageDependencyDef.
        Column data-dependencies in format: output_column => list of its input columns

        :return: The column_data_dependencies of this SchemaLineageDependencyDef.
        :rtype: dict(str, list[str])
        """
        return self._column_data_dependencies

    @column_data_dependencies.setter
    def column_data_dependencies(self, column_data_dependencies):
        """
        Sets the column_data_dependencies of this SchemaLineageDependencyDef.
        Column data-dependencies in format: output_column => list of its input columns

        :param column_data_dependencies: The column_data_dependencies of this SchemaLineageDependencyDef.
        :type: dict(str, list[str])
        """

        self._column_data_dependencies = column_data_dependencies

    @property
    def column_control_dependencies(self):
        """
        Gets the column_control_dependencies of this SchemaLineageDependencyDef.
        Column control-dependencies in format: output_column => list of its input columns

        :return: The column_control_dependencies of this SchemaLineageDependencyDef.
        :rtype: dict(str, list[str])
        """
        return self._column_control_dependencies

    @column_control_dependencies.setter
    def column_control_dependencies(self, column_control_dependencies):
        """
        Sets the column_control_dependencies of this SchemaLineageDependencyDef.
        Column control-dependencies in format: output_column => list of its input columns

        :param column_control_dependencies: The column_control_dependencies of this SchemaLineageDependencyDef.
        :type: dict(str, list[str])
        """

        self._column_control_dependencies = column_control_dependencies

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, SchemaLineageDependencyDef):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



import logging
from hashlib import sha256

from kensu.client import UserPK, CodeBasePK, CodeVersionPK , PhysicalLocationPK , DataSourcePK , ProcessPK , SchemaPK , \
                                ProcessLineagePK , ProcessRunPK , LineageRunPK , DataStatsPK , ModelPK , ModelTrainingPK , \
                                ModelMetricsPK  , ProcessRunStatsPK , ProjectPK



class KensuClassHandlers(object):

    @classmethod
    def guid_pk(cls, pk):
        sPK = cls.serializePK(pk)
        hashed = sha256(sPK.encode("utf-8")).hexdigest()
        return "k-" + hashed

    @classmethod
    def guid(cls, o):
        return cls.guid_pk(o.pk)

    # Case switcher to create main map of fields => see `serializePK` function
    @classmethod
    def caseUserPK(cls, o):
        return {
            "_dam_entity_type": "USER"
            , "name": o.name
        }

    @classmethod
    def caseCodeBasePK(cls, o):
        return {
            "_dam_entity_type": "CODE_BASE"
            , "location": o.location
        }

    @classmethod
    def caseCodeVersionPK(cls, o):
        return {
            "_dam_entity_type": "CODE_VERSION"
            , "version": o.version
            , "codebaseRef": cls.serializeRef(o.codebase_ref)
        }

    @classmethod
    def casePhysicalLocationPK(cls, o):
        return {
            "_dam_entity_type": "PHYSICAL_LOCATION"
            , "city": o.city.lower()
            , "country": o.country.lower()
        }

    @classmethod
    def caseDataSourcePK(cls, o):
        return {
            "_dam_entity_type": "DATA_SOURCE"
            , "location": o.location
            , "physical-location": cls.serializeRef(o.physical_location_ref)
        }

    @classmethod
    def caseProcessPK(cls, o):
        return {
            "_dam_entity_type": "PROCESS"
            , "qualifiedName": o.qualified_name
        }

    @classmethod
    def caseSchemaPK(cls, o):
        return {
            "_dam_entity_type": "SCHEMA"
            , "dataSourceRef": cls.serializeRef(o.data_source_ref)
            , "fields": cls.serializeFields(o.fields)
        }

    @classmethod
    def caseProcessLineagePK(cls, o):
        return {
            "_dam_entity_type": "PROCESS_LINEAGE"
            , "processRef": cls.serializeRef(o.process_ref)
            , "dataFlow": cls.serializeDataFlow(o.data_flow)
        }

    @classmethod
    def caseProcessRunPK(cls, o):
        return {
            "_dam_entity_type": "PROCESS_RUN"
            , "processRef": cls.serializeRef(o.process_ref)
            , "qualifiedName": o.qualified_name
        }

    @classmethod
    def caseLineageRunPK(cls, o):
        return {
            "_dam_entity_type": "LINEAGE_RUN"
            , "lineageRef": cls.serializeRef(o.lineage_ref)
            , "processRunRef": cls.serializeRef(o.process_run_ref)
            , "timestamp": o.timestamp
        }

    @classmethod
    def caseDataStatsPK(cls, o):
        return {
            "_dam_entity_type": "DATA_STATS"
            , "schemaRef": cls.serializeRef(o.schema_ref)
            , "lineageRunRef": cls.serializeRef(o.lineage_run_ref)
        }

    @classmethod
    def caseModelPK(cls, o):
        return {
            "_dam_entity_type": "MODEL"
            , "name": o.name
        }

    @classmethod
    def caseModelTrainingPK(cls, o):
        return {
            "_dam_entity_type": "MODEL_TRAINING"
            , "modelRef": cls.serializeRef(o.model_ref)
            , "processLineageRef": cls.serializeRef(o.process_lineage_ref)
        }

    @classmethod
    def caseModelMetricsPK(cls, o):
        return {
            "_dam_entity_type": "MODEL_METRICS"
            , "modelTrainingRef": cls.serializeRef(o.model_training_ref)
            , "lineageRunRef": cls.serializeRef(o.lineage_run_ref)
            , "storedInSchemaRef": cls.serializeRef(o.stored_in_schema_ref)
        }

    @classmethod
    def caseModelMonitorPK(cls, o):
        return {
            "_dam_entity_type": "MODEL_MONITOR"
            , "monitoredSchemaRef": cls.serializeRef(o.monitored_schema_ref)
            , "timestamp": o.timestamp
        }

    @classmethod
    def caseProcessRunStatsPK(cls, o):
        return {
            "_dam_entity_type": "PROCESS_RUN_STATS"
            , "processRunRef": cls.serializeRef(o.process_run_ref)
            , "timestamp": o.timestamp
        }

    @classmethod
    def caseProjectPK(cls, o):
        return {
            "_dam_entity_type": "PROJECT"
            , "name": o.name
        }

    @classmethod
    def caseSchemaFieldTagPK(cls, o):
        return {
            "_dam_entity_type": "SCHEMA_FIELD_TAG"
            , "schemaRef": cls.serializeRef(o.schema_ref)
            , "fieldName": o.field_name
            , "tagID": o.tag_id
        }

    CaseSwitcher = {
        UserPK: caseUserPK,
        CodeBasePK: caseCodeBasePK,
        CodeVersionPK: caseCodeVersionPK,
        PhysicalLocationPK: casePhysicalLocationPK,
        DataSourcePK: caseDataSourcePK,
        ProcessPK: caseProcessPK,
        SchemaPK: caseSchemaPK,
        ProcessLineagePK: caseProcessLineagePK,
        ProcessRunPK: caseProcessRunPK,
        LineageRunPK: caseLineageRunPK,
        DataStatsPK: caseDataStatsPK,
        ModelPK: caseModelPK,
        ModelTrainingPK: caseModelTrainingPK,
        ModelMetricsPK: caseModelMetricsPK,
        ProcessRunStatsPK: caseProcessRunStatsPK,
        ProjectPK: caseProjectPK,
    }

    @classmethod
    def serializePK(cls, o):
        clazz = o.__class__
        attrs = clazz.swagger_types

        keys = cls.CaseSwitcher[clazz].__func__(KensuClassHandlers, o)

        if (len(keys) != len(attrs) + 1):  # +1 because we add `_kensu_entity_type`
            logging.error(
                "Error serialization keys, it looks like there are different counts " + 
                "of serialized keys than fields in the pk of clazz=" +
                str(clazz) +
                ", attrs=", str(attrs) +
                ", keys=", str(keys)
            )

        serializedKeys = [str.format("{} -> {}", k, v) for [k, v] in keys.items()]
        serializedKeys.sort()
        serializedPK = "\n".join(serializedKeys)
        return serializedPK

    @classmethod
    def serializeRef(cls, ref):
        if ref.by_guid is not None:
            return ref.by_guid
        else:
            return cls.guid_pk(ref.by_pk)

    @classmethod
    def serializeFields(cls, fields):
        fieldsAsStrings = [
            "FieldDef(name: {}, fieldType: {}, nullable: {})".format(f.name, f.field_type, str(f.nullable).lower()) for
            f in fields]
        fieldsAsStrings.sort()
        serializedFields = "[" + ",".join(fieldsAsStrings) + "]"
        return serializedFields

    @classmethod
    def serializeDataFlow(cls, defs):
        def handleDef(d):
            fr = cls.serializeRef(d.from_schema_ref)
            to = cls.serializeRef(d.to_schema_ref)
            cdd = cls.serializeDependencies(d.column_data_dependencies)
            ccd = cls.serializeDependencies(d.column_control_dependencies)
            return "SchemaLineageDependencyDef(from: {}, to: {}, dataDeps: {}, controlDeps: {})".format(fr, to, cdd,
                                                                                                        ccd)

        defsAsStrings = [handleDef(d) for d in defs]
        defsAsStrings.sort()
        serializedDataFlow = "[" + ",".join(defsAsStrings) + "]"
        return serializedDataFlow

    @classmethod
    def serializeDependencies(cls, deps):
        if deps is None or len(deps) == 0:
            return "{}"
        else:
            depsAsStrings = ["{} -> [{}]".format(outputField, ",".join(sorted(inputFields))) for
                             [outputField, inputFields] in deps.items()]
            depsAsStrings.sort()

            serializedDeps = "{" + ",".join(depsAsStrings) + "}"
            return serializedDeps
